/*== Link include files ==============================================*/

#include <dos.h>
#include <stdio.h>
#include <string.h>

#define SEG( p ) ( ( unsigned int ) ( ( ( long ) p ) >> 16 ) )
#define OFS( p ) ( ( unsigned int ) ( p ) )

/*== Constants =======================================================*/

#define NOPE         0x4E                                 /* N for No */
#define NO_DRIVE     0                        /* Drive does not exist */
#define DD_525       1                             /* Drive: 5.25" DD */
#define HD_525       2                             /* Drive: 5.25" HD */
#define DD_35        3                              /* Drive: 3.5" DD */
#define HD_35        4                              /* Drive: 3.5" HD */
#define MAXNUMTRIES  5                     /* Maximum number of tries */

#define ReadOp  0x02                     /* Read Operation */
#define WriteOp 0x03                     /* Write Operation */

#define TRUE  ( 1 == 1 )
#define FALSE ( 1 == 0 )

/*== Macros ======================================================*/

#define HI(x) ( *((BYTE *) (&x)+1) ) /* Returns high byte of word */
#define LO(x) ( *((BYTE *) &x) )      /* Returns low byte of word */


/*== Type declarations ===========================================*/

typedef unsigned char BYTE;                         /* Data type byte */
typedef unsigned int WORD;

typedef BYTE TrackBfType[ 18 ][ 512 ];          /* Memory for a track */

typedef struct {	                  /* Floppy disk drive parameters */
		  char    *stype;
		  BYTE	  DSides,                  /* Number of sides */
			  STrax,                   /* Tracks           */
			  TSectors;                /* Sectors per track */

		} FloppyType;


typedef struct {
		  BYTE    HardNo,                   /* The number of Hards */
			  RWHeads;                  /* Number of R/W Heads */
		  int	  Celenders,                   /* Celenders        */
			  TSectors;                /* Sectors per track */

		} HardType;


typedef struct {                   /* Describes a sector position */
		  BYTE Head;                   /* Read/write head */
		  WORD SecCyl;      /* Sector and cylinder number */
		} SECPOS;

typedef struct {                         /* Partition table entry */
		  BYTE          Status;       /* Partition status */
		  SECPOS        StartSec;         /* First sector */
		  BYTE          PartTyp;        /* Partition type */
		  SECPOS        EndSec;            /* Last sector */
		  unsigned long SecOfs;  /* Offset of boot sector */
		  unsigned long SecNum;      /* Number of sectors */
		} PARTENTRY;

typedef struct {                    /* Describes partition sector */
		  BYTE      BootCode[ 0x1BE ];
		  PARTENTRY PartTable[ 4 ];
		  WORD      IdCode;                     /* 0xAA55 */
		} PARTSEC;

typedef PARTSEC far *PARSPTR;
			    /* Ptr. to partition sector in memory */

typedef BYTE SectorSize[ 512 ] ;          /* Memory for a Sector */
typedef SectorSize far *SectorSizeP; /*Ptr. to a sector */

/**********************************************************************/
/* ResetDisk : Disk reset on all drives.                              */
/* Input     : None                                                   */
/* Output    : None                                                   */
/* Info      : Reset executed on all drives, regardless of drive      */
/*             number loaded in DD                                    */
/**********************************************************************/

void DiskReset( void )
{
 union REGS regs;           /* Processor registers for interrupt call */

 regs.h.ah = 0x00;              /* Function number for interrupt call */
 regs.h.dl = 0;                                /* Drive a: (see Info) */
 int86( 0x13, &regs, &regs );                       /* Interrupt call */
}

/******************************************************************/
/*  GetSecCyl: Gets the combined sector/cylinder coding of the    */
/*             BIOS sector and cylinder number.                   */
/*  Input  : SecCyl   : Value to be decoded                       */
/*           Sector   : Sector variable reference                 */
/*           Cylinder : Cylinder variable reference               */
/*  Output : None                                                 */
/******************************************************************/

void GetSecCyl( WORD SecCyl, int *Sector, int *Cylinder )

{

 *Sector   = SecCyl & 63;                    /* Mask bits 6 and 7 */
 *Cylinder = HI(SecCyl) + ( ( (WORD) LO(SecCyl) & 192 ) << 2 );
}

/**********************************************************************/
/* GetFloppyType: Determines the Floppy drive type.                   */
/* Input        : DRIVE = Drive number                                */
/* Output       : Floppy spcefications as a struct of FloppyType      */
/**********************************************************************/

FloppyType GetFloppyType (BYTE Drive)
{
 union REGS regs;	   /* Processor registers for interrupt call */
 FloppyType     info;
 int type=0;
 regs.h.ah = 0x08;                  /* Function: Determine drive type */
 regs.h.dl = Drive;                                   /* Drive number */
 int86( 0x13, &regs, &regs );                  /* Call BIOS interrupt */
 info.DSides= regs.h.dh;         /*save the sides number*/
 info.STrax= regs.h.ch;         /*save the tracks number*/
 info.TSectors= regs.h.cl;      /*save the sectors number*/
 type= regs.h.bl;
 switch (type)
 {
 case NOPE:info.stype="Nope";
	   break;
 case NO_DRIVE:info.stype="Drive does not exist";
	       break;
 case DD_525:info.stype="Drive: 5.25 DD"; /*5.25 Double Density*/
	   break;
 case HD_525:info.stype="Drive: 5.25 HD"; /*5.25 High Density*/
	   break;
 case DD_35:info.stype="Drive: 3.5 DD";   /*3.5 Double Density*/
	   break;
 case HD_35:info.stype="Drive: 3.5 HD";   /*3.5 High Density*/
 }

 return (info);
}

/**********************************************************************/
/* GetHardType: Determines the Hard drive type.                       */
/* Input        : DRIVE = Drive number                                */
/* Output       : Hard spcefications as a struct of HardType          */
/**********************************************************************/
HardType GetHardType (BYTE Drive)
{
 union REGS regs;
 HardType    Hinfo;
 WORD cylcsec;
 regs.h.ah = 0x08;                  /* Function: Determine drive type */
 regs.h.dl = Drive;                                   /* Drive number */
 int86( 0x13, &regs, &regs );                  /* Call BIOS interrupt */
 Hinfo.HardNo= regs.h.dl;         			   /*save hards number*/
 Hinfo.RWHeads=regs.h.dh+1;       			   /*save read write heads number*/
 GetSecCyl( regs.x.cx, &Hinfo.TSectors, &Hinfo.Celenders); /*save cylinders number*/
 return (Hinfo);
}



/******************************************************************/
/*ReadPartSec     :Reads a partition sector from the hard drive.  */
/*Input  :- DS    : BIOS code for drive (0x80, 0x81, etc.)        */
/*        - Head  : Number of read/write heads                    */
/*        - SctCyl: Sector/cylinder numbers in BIOS format        */
/*        - Buf   : Buffer to which sector is passed              */
/*Output : TRUE if sector can be read without errors,             */
/*         otherwise FALSE                                      */
/******************************************************************/

BYTE ReadPartSec( BYTE Drive, BYTE Head, WORD SecCyl, PARSPTR Buf )

{
 union REGS   Regs;     /* Processor registers for interrupt call */
 struct SREGS SRegs;

 Regs.x.ax = 0x0201;         /* Funct. no.: READ for first sector */
 Regs.h.dl = Drive;                      /* Pass other parameters */
 Regs.h.dh = Head;                       /* to their respective   */
 Regs.x.cx = SecCyl;                     /* registers             */
 Regs.x.bx = FP_OFF( Buf );
 SRegs.es  = FP_SEG( Buf );

 int86x( 0x13, &Regs, &Regs, &SRegs );
				     /* Call hard drive interrupt */
 return !Regs.x.cflag;
}

/**********************************************************************/
/*ReadWriteHardSec : Reads or Writes a sector from\to the hard drive. */
/*Input  : - DS    : BIOS code for drive (0x80, 0x81, etc.)           */
/*         - OpNum : Opperation Number 0x02:Read,0x03:Write        */
/*         - Sec   : Number of Sector                                 */
/*         - Head  : Number of read/write heads                       */
/*         - SctCyl: Sector/cylinder numbers in BIOS format           */
/*         - Buf   : Buffer to which sector is passed                 */
/*Output : TRUE if sector can be read without errors,                 */
/*         otherwise FALSE                                            */
/**********************************************************************/

BYTE ReadWriteHardSec( BYTE Drive,BYTE OpNum,BYTE Sec ,BYTE Head, WORD SecCyl,SectorSizeP Buf )

{
 union REGS   Regs;     /* Processor registers for interrupt call */
 struct SREGS SRegs;

 Regs.h.ah = OpNum;       /* Funct. no.: READ\Write sector    */
 Regs.h.al = Sec;         	     /* Pass other parameters */
 Regs.h.dl = Drive;       	     /* to their respective   */
 Regs.h.dh = Head;
 Regs.x.cx = SecCyl;
 Regs.x.bx = FP_OFF( Buf ); 	     /* registers             */
 SRegs.es  = FP_SEG( Buf );

 int86x( 0x13, &Regs, &Regs, &SRegs );
				     /* Call hard drive interrupt */
 return !Regs.x.cflag;

}

/**************************************************************************/
/*ReadWriteFloppySec : Reads or Writes a sector from\to the Floppy drive. */
/*Input  : - DS      : BIOS code for drive (0x00, 0x01, etc.)             */
/*         - OpNum   : Opperation Number 0x02:Read,0x03:Write             */
/*         - Side    : Number of Side 1st=0x00 ,2nd=0x01   	          */
/*         - Sec     : Number of Sector                                   */
/*         - Track   : Number of Track                                    */
/*         - Buf     : Buffer to which sector is passed                   */
/*Output : TRUE if sector can be read without errors,                     */
/*         otherwise FALSE                                                */
/**************************************************************************/

BYTE ReadWriteFloppySec( BYTE Drive,BYTE OpNum,BYTE Side ,BYTE Sec,BYTE Track,SectorSize* Buf)

{
 union REGS   Regs;     /* Processor registers for interrupt call */
 struct SREGS SRegs;

 Regs.h.ah=OpNum;         /* Funct. no.: READ\Write sector        */
 Regs.h.al=1;                   /*Reading\writing only one sector */
 Regs.h.dl=Drive;               /* Pass other parameters          */
 Regs.h.dh=Side;                /* to their respective            */
 Regs.h.cl=Sec;
 Regs.h.ch=Track;
 Regs.x.bx = OFS( *Buf );                  /* Offset addr. of buffer */
 SRegs.es = SEG( *Buf );                   /* Segment addr.          */
 int86x( 0x13, &Regs, &Regs, &SRegs );     /* Call BIOS interrupt */
 return !Regs.x.cflag;
}


/******************************************************************/
/*  ShowPartition: Displays hard drive partitioning on the screen.*/
/*  Input  : DS : Number of the corresponding hard drive          */
/*                (0, 1, etc.)                                    */
/*  Output : None                                                 */
/******************************************************************/

void ShowPartition( BYTE LW )
{
 #define AP ( ParSec.PartTable[ Entry ] )

 BYTE       Head,                    /* Head of current partition */
	    Entry;                                /* Loop counter */
 int        Sector,                            /* Get sector and  */
	    Cylinder;                          /* cylinder number */
 PARTSEC    ParSec;                   /* Current partition sector */
 union REGS Regs;       /* Processor registers for interrupt call */

 printf("\n");
 LW |= 0x80;                     /* Prepare drive number for BIOS */
 if ( ReadPartSec( LW, 0, 1, &ParSec ) ) /* Read partition sector */
  {                                       /* Sector could be read */
   Regs.h.ah = 8;                    /* Funct. no.: Read drive ID */
   Regs.h.dl = LW;
   int86( 0x13, &Regs, &Regs );      /* Call hard drive interrupt */
   GetSecCyl( Regs.x.cx, &Sector, &Cylinder );
   printf( "+-------------------------------------------"
	    "------------------------------+\n");
   printf( "      Drive %2d:     %2d Heads %4d"
	    " Cylinders %3d Sectors                            \n",
	    LW-0x80, Regs.h.dh+1, Cylinder, Sector );
   printf( "      Partition table in Partition Sector              \n");
   printf( "---------------------------------------------"
	   "------------------------------\n");
   printf( "                            Start          End"
	   "            Dist fm          \n");
   printf( "Nr Boot Type                Head Cyl. Sec  Head"
	   " Cyl. Sec BootSec Total\n");
   printf( "--+----+-------------------+--------------+----"
	   "----------+-------+---------\n");

   /*-- Get partition table --------------------------------------*/
   for ( Entry=0; Entry < 4; ++Entry )
    {
     printf( " %d", Entry );
     if ( AP.Status == 0x80 )                /* Partition active? */
      printf("  Yes");
     else
      printf("  No ");
     printf(" ");
     switch( AP.PartTyp )               /* Compute partition type */
      {
       case 0x00 : printf( "Not allocated      " );
	       break;
       case 0x01 : printf( "DOS, 12-bit FAT    " );
	       break;
       case 0x02 :
       case 0x03 : printf( "XENIX              " );
	       break;
       case 0x04 : printf( "DOS, 16-bit FAT    " );
	       break;
       case 0x05 : printf( "DOS, ext. partition" );
	       break;
       case 0x06 : printf( "DOS 4.0 > 32 MB    " );
	       break;
       case 0xDB : printf( "Concurrent DOS     " );
	       break;
       default   : printf( "Unknown (%3d)      ",
		     ParSec.PartTable[ Entry ].PartTyp );
      }

     /*-- Get physical and logical parameters --------------------*/
     GetSecCyl( AP.StartSec.SecCyl, &Sector, &Cylinder );
     printf( "%2d %5d  %3d    ", AP.StartSec.Head, Cylinder, Sector );
     GetSecCyl( AP.EndSec.SecCyl, &Sector, &Cylinder );
     printf( "%2d %5d  %3d  ", AP.EndSec.Head, Cylinder, Sector );
     printf( "%6lu %6lu  \n", AP.SecOfs, AP.SecNum );
    }
   printf( "+-------------------------------------------"
	    "------------------------------+\n" );
  }
 else
  printf("Error during boot sector access\n");
}


/**********************************************************************/
/*                         MAIN PROGRAM                               */
/**********************************************************************/

void main()
{
 char c,again='n';
 FloppyType floppyinfo;
 HardType       hinfo;
 int a1,b1,c1,d1;
 int i=0,length1,from1;

 SectorSize SectorData;

 do
 {
   clrscr();
   printf("\nSelect the drive you want to deal with :");
   printf("\n1) Floppy Drive");
   printf("\n2) Hard Drive   \"doesn't work on WinXP\"\n");
   do
   {
     scanf("%c",&c);
   }while((c!='1')&&(c!='2'));
   if (c=='1')          /*Floppy Drive Operations*/
   {
	   printf("Select one task of the following to be performed on the Floppy drive:");
	   printf("\n1) Get properties of the floppy disk");
	   printf("\n2) Read a Sector from the floppy disk");
	   printf("\n3) Write a Sector to the floppy disk\n");
	   do
	   {
	     scanf("%c",&c);
	   }while((c!='1')&&(c!='2')&&(c!='3'));
	 switch (c)
	 {
	   case '1':      /*properties of the floppy disk*/
		     floppyinfo=GetFloppyType (0);
		     printf("\n\n The Type of Floppy drive is %s",floppyinfo.stype);
		     printf("\n The number of Sides %d",floppyinfo.DSides);
		     printf("\n The number of Tracks %d",floppyinfo.STrax+1);
		     printf("\n The number of Sectors %d",floppyinfo.TSectors);
	   break;
	   case '2':      /*Read a Sector from the floppy disk*/
		     printf("\nEnter Side Number\n");
		     scanf("%d",&a1);
		     printf("Enter Sector Number\n");
		     scanf("%d",&b1);
		     printf("Enter Track Number\n");
		     scanf("%d",&c1);
		     if (ReadWriteFloppySec( 0x00,ReadOp,a1,b1,c1,&SectorData))
		     {
			i=0;
			while (i<512)                    /*Printing the Sector*/
			{                                      /*..*/
			   if (i%25==0)                        /*..*/
			     printf("\n");                     /*..*/
			   printf("%3x",SectorData[i]);        /*..*/
			   i++;                                /*..*/
			}

		     }
		     else
		     {
			    printf("Operation Failed");   /*Error in Reading*/
			    DiskReset();
		     }
	   break;
	   case '3': /*Write a Sector to the floppy disk*/
		     printf("\nEnter Side Number\n");
		     scanf("%d",&a1);
		     printf("Enter Sector Number\n");
		     scanf("%d",&b1);
		     printf("Enter Track Number\n");
		     scanf("%d",&c1);
		     if (!ReadWriteFloppySec( 0x00,ReadOp,a1,b1,c1,&SectorData))/*Read the Sector Bytes*/
		     {
			printf("Operation Failed\n");  /*Error in Reading*/
			DiskReset();
			break;
		     }
		     printf("Enter How Many Bytes you want to write\n");
		     scanf("%d",&length1);
		     printf("Enter the number of byte you want to start from\n");
		     scanf("%d",&from1);

		     if ((length1<0)||(from1<0)||((length1+from1)>=512)) /*Invalid Inputs*/
		     {
			printf("Operation Failed\n");  /*Can't perform the writing operation*/
			break;
		     }

		     i=from1;                          /*modifiy the Sector Bytes*/
		     while (i<(from1+length1))
		     {
		       printf("Enter %d byte: ",i-from1);
		       scanf("%x",&SectorData[i]);
		       i++;
		     }

		     if (ReadWriteFloppySec( 0x00,WriteOp,a1,b1,c1,&SectorData)) /*Write the Sector after modification*/
		     {
			    printf("Operation Done!");
		     }
		     else
		     {
			    printf("Operation Failed");   /*Error in Writing*/
			    DiskReset();
		     }

	 }
   }
   else            /*Hard Drive Operations*/
   {
	   printf("Select one task of the following to be performed on the Hard drive:");
	   printf("\n1) Get properties of the Hard disk");
	   printf("\n2) Read a Sector from the hard disk ");
	   printf("\n3) Show partion sector properties \n ");
	   do
	   {
	     scanf("%c",&c);
	   }while((c!='1')&&(c!='2')&&(c!='3'));
	 switch (c)
	 {
	   case '1':       /*Get properties of the Hard disk*/
		     hinfo=GetHardType (0x80);
		     printf("\n\n The Number of hard drives connected to the controller %d",hinfo.HardNo);
		     printf("\n The number of read/write heads %d",hinfo.RWHeads);
		     printf("\n The number of Celenders %4d",hinfo.Celenders);
		     printf("\n The number of Sectors %3d",hinfo.TSectors);
	   break;
	   case '2':       /*Read a Sector from the hard disk*/
		     printf("\nEnter Sector Number\n");
		     scanf("%d",&a1);
		     printf("Enter Head Number\n");
		     scanf("%d",&b1);
		     printf("Enter Cylinder Number\n");
		     scanf("%d",&c1);
		     if (ReadWriteHardSec( 0x80,ReadOp,a1,b1,c1,&SectorData))
		     {
			i=0;
			while (i<512)                         /*Printing the Sector*/
			{                                            /*..*/
			   if (i%25==0)                              /*..*/
			     printf("\n");                           /*..*/
			   printf("%3x",SectorData[i]);              /*..*/
			   i++;                                      /*..*/
			}                                            /*..*/

		     }
		     else
		     {
			    printf("Operation Failed");     /*Error in Reading*/
			    DiskReset();
		     }
	   break;
	   case '3':
		     ShowPartition(0x80);       /* Display partition sector */
	 }

   }
 printf("\n\n Do you want to do another operation? Y/yes\n");  /*Asking to do again?*/
 scanf("%s",&again);
 }while((again=='y')||(again=='Y'));     /*if again is not equal to ('Y'or'y') then Exit*/
}